#include <stdio.h>
#include <math.h>


static double *_points = NULL;
static double *_pixels = NULL;
static double *_maxParmValues = NULL;
static int _numPts = -1;
static int _numParms = -1;



extern "C" int lmdif_(int (*)(int *, int *, double *, double *), 
                      int *m, int *n, double *x, double *fvec, 
                      double *ftol, double *xtol, double *gtol, 
                      int *maxfev, double *epsfcn, double *diag, 
                      int *mode, double *factor, int *nprint, 
                      int *info, int *nfev, double *fjac, 
                      int *ldfjac, int *ipvt, double *qtf, 
                      double *wa1, double *wa2, double *wa3, 
                      double *wa4);


/*****************************************************************************
*****************************************************************************/
static int
lmdifError (int *m_ptr, int *n_ptr, double *parms, double *error)
{
   int nparms = *n_ptr;
   int nerrors = *m_ptr;

   assert(nparms == _numParms);
   assert(nerrors == _numPts);


   double f  = parms[0] * _maxParmValues[0];
   double cx = parms[1] * _maxParmValues[1];
   ...

   int count = 0;
   for (int p=0; p<nerrors; p++) {
      error[count++] = ....;
   }

   return (TRUE);
}


/*****************************************************************************
*****************************************************************************/
/* Parameters controlling MINPACK's lmdif() optimization routine. */
/* See the file lmdif.f for definitions of each parameter.        */
#define REL_SENSOR_TOLERANCE_ftol    1.0E-5      /* [pix] */
#define REL_PARAM_TOLERANCE_xtol     1.0E-7
#define ORTHO_TOLERANCE_gtol         0.0
#define MAXFEV                       (1000*n)
#define EPSFCN                       1.0E-6 /* was E-16 Do not set to 0! */
#define MODE                         2      /* variables scaled here */
#define FACTOR                       1.0 
double
nonLinearLeastSquares(double *points, double *pixels, int numPts, 
					  double *cameraParms, double *maxParmValues, int numParms)
					 
{
    /* Parameters needed by MINPACK's lmdif() */
    double *x;
    double *fvec;
    double  ftol = REL_SENSOR_TOLERANCE_ftol;
    double  xtol = REL_PARAM_TOLERANCE_xtol;
    double  gtol = ORTHO_TOLERANCE_gtol;
    double *diag;
    int     mode = MODE;
    double  factor = FACTOR;
	double  epsfcn = EPSFCN;
    int     nprint = 0;
    int     info;
    int     nfev;
    double *fjac;
    int    *ipvt;
    double *qtf;
    double *wa1;
    double *wa2;
    double *wa3;
    double *wa4;

	/* set globals to that callback can get to data */
	_points = points;
	_pixels = pixels;
	_maxParmValues = maxParmValues;
	_numPts = numPts;
	_numParms = numParms;

	 /* set m and n */
    int m = numPts; // num equations
	int n = numParms; // num parms to optimize
    int ldfjac = m;
    int maxfev = MAXFEV;

    /* allocate stuff dependent on n */
    x    = (double *)calloc(n, sizeof(double));
    diag = (double *)calloc(n, sizeof(double));
    qtf  = (double *)calloc(n, sizeof(double));
    wa1  = (double *)calloc(n, sizeof(double));
    wa2  = (double *)calloc(n, sizeof(double));
    wa3  = (double *)calloc(n, sizeof(double));
    ipvt = (int    *)calloc(n, sizeof(int));


    /* allocate some workspace */
    if (( fvec = (double *) calloc ((unsigned int) m, 
                                    (unsigned int) sizeof(double))) == NULL ) {
       fprintf(stderr,"calloc: Cannot allocate workspace fvec\n");
       exit(-1);
    }

    if (( fjac = (double *) calloc ((unsigned int) m*n,
                                    (unsigned int) sizeof(double))) == NULL ) {
       fprintf(stderr,"calloc: Cannot allocate workspace fjac\n");
       exit(-1);
    }

    if (( wa4 = (double *) calloc ((unsigned int) m, 
                                   (unsigned int) sizeof(double))) == NULL ) {
       fprintf(stderr,"calloc: Cannot allocate workspace wa4\n");
       exit(-1);
    }


    /* copy parameters in as initial values */
    for (int i=0; i<n; i++)
       x[i] = cameraParms[i] / maxParmValues[i];

	    /* define optional scale factors for parameters (only used if mode==2) */
    if ( mode == 2 ) {
		 int offset = 0;
		 for (int offset=0; offset<numParms; offset++) {
            diag[offset] = 1.0;
		 } 
    }

    /* perform the optimization */ 
	 if (verbose)
       printf("Starting optimization step (n=%i, m=%i)...\n", n, m);
    lmdif_ (lmdifError,
            &m, &n, x, fvec, &ftol, &xtol, &gtol, &maxfev, &epsfcn,
            diag, &mode, &factor, &nprint, &info, &nfev, fjac, &ldfjac,
            ipvt, qtf, wa1, wa2, wa3, wa4);
    double totalerror = 0;
    for (i=0; i<m; i++)
       totalerror += fvec[i];
	 if (verbose) {
       printf("\tnum function calls = %i\n", nfev);
       printf("\tinfo = %i\n", info);
       printf("\tremaining total error value = %f\n", totalerror);
       printf("\tor %1.2f mm per point\n", SQRT(totalerror / m));
       printf("...ended optimization step.\n");
	 }

	 /* return the error */
	 retErr = SQRT(totalerror / m);

    /* copy result back to parameters array */
    for (i=0; i<n; i++)
       cameraParms[i] = x[i] * maxParmValues[i];

    /* release allocated workspace */
    free (fvec);
    free (fjac);
    free (wa4);
    free (ipvt);
    free (wa1);
    free (wa2);
    free (wa3);
    free (qtf);
    free (diag);
    free (x);

	return (retErr);
}
